# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from collections import OrderedDict
import inspect
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_base import ParallelMaster, JobGenerator, get_function_from_string
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class AtomisticParallelMaster(ParallelMaster, AtomisticGenericJob):
    def __init__(self, project, job_name):
        super(AtomisticParallelMaster, self).__init__(project, job_name=job_name)

    @property
    def structure(self):
        if self.ref_job is not None:
            return self._ref_job.structure
        else:
            return None

    @structure.setter
    def structure(self, basis):
        if self.ref_job is not None:
            self._ref_job.structure = basis
        else:
            raise ValueError(
                "A structure can only be set after a reference job has been assinged."
            )

    def _get_structure(self, frame=0, wrap_atoms=True):
        return self.structure

    def _number_of_structures(self):
        return 1


class GenericOutput(OrderedDict):
    def __init__(self):
        super(GenericOutput, self).__init__()


class MapMaster(AtomisticParallelMaster):
    """
    Atomistic parallel master job that computes a given function in the given arguments in parallel.

    Example usage:

    >>> master = self.create_job(
    ...     job_type=self.project.job_type.MapMaster, job_name="map_" + self.job_name
    ... )
    >>> master.modify_function = function
    >>> master.parameter_list = parameter_lst

    or simpler

    >>> master = job.map(function, parameter_lst)


    """

    def __init__(self, project, job_name):
        """

        Args:
            project:
            job_name:
        """
        super(MapMaster, self).__init__(project, job_name)

        self.__version__ = "0.0.1"
        self._job_generator = MapJobGenerator(self)
        self._map_function = None
        self.parameter_list = []

    @property
    def modify_function(self):
        return self._map_function

    @modify_function.setter
    def modify_function(self, funct):
        self._map_function = funct

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the ParameterMaster in an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(MapMaster, self).to_hdf(hdf=hdf, group_name=group_name)
        if len(self.parameter_list) != 0:
            with self.project_hdf5.open("input") as hdf5_input:
                first_element = self.parameter_list[0]
                if isinstance(first_element, Atoms):
                    with hdf5_input.open("structures") as hdf5_input_str:
                        for ind, struct in enumerate(self.parameter_list):
                            struct.to_hdf(
                                hdf=hdf5_input_str, group_name="s_" + str(ind)
                            )
                elif isinstance(first_element, (int, float, str, list)):
                    hdf5_input["parameters_list"] = self.parameter_list
                else:
                    raise TypeError()
                if self._map_function is not None:
                    try:
                        hdf5_input["map_function"] = inspect.getsource(
                            self._map_function
                        )
                    except IOError:
                        hdf5_input["map_function"] = "None"
                else:
                    hdf5_input["map_function"] = "None"

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the ParameterMaster from an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(MapMaster, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            if "structures" in hdf5_input.list_groups():
                with hdf5_input.open("structures") as hdf5_input_str:
                    self.parameter_list = [
                        Atoms().from_hdf(hdf5_input_str, group_name)
                        for group_name in sorted(hdf5_input_str.list_groups())
                    ]
            else:
                self.parameter_list = hdf5_input["parameters_list"]
            function_str = hdf5_input["map_function"]
            if function_str == "None":
                self._map_function = None
            else:
                self._map_function = get_function_from_string(function_str)

    def collect_output(self):
        pass


class MapJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        return self._master.parameter_list

    def modify_job(self, job, parameter):
        return self._master._map_function(job, parameter)


def pipe(project, job, step_lst, delete_existing_job=False):
    """
    Create a job pipeline

    Args:
        project (pyiron.Project): Project to calculate pipeline in
        job (AtomisticGenericJob): Template for the calculation
        step_lst (list): List of functions which create calculations

    Returns:
        FlexibleMaster:
    """
    job_lst_master = project.create_job(
        project.job_type.FlexibleMaster,
        job.job_name + "_lstmaster",
        delete_existing_job=delete_existing_job,
    )
    if job_lst_master.status.finished:
        return job_lst_master
    else:
        for i, step_funct in enumerate(step_lst):
            job_lst_master.append(step_funct(job))
            if i > 0 and "for_each_structure" in step_funct.__name__:
                job_lst_master.function_lst.append(_structure_many_to_many)
            elif i > 0:
                job_lst_master.function_lst.append(_structure_one_to_one)
        return job_lst_master


def _structure_one_to_one(job_prev, job_next):
    """
    Copy final structure from previous job to the input structure of the next job

    Args:
        job_prev (AtomisticGenericJob): previous Job object
        job_next (AtomisticGenericJob): next Job object
    """
    job_next.structure = job_prev.get_structure(iteration_step=-1)


def _structure_many_to_many(job_prev, job_next):
    """
    Copy list of final structures from previous job to the input structure list of the next job

    Args:
        job_prev (AtomisticParallelMaster): previous Job object
        job_next (AtomisticParallelMaster): next Job object
    """
    job_next.structure_lst = job_prev.list_structures()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
import os
import scipy.constants
import subprocess
import warnings
import time
from pyiron_atomistics.sphinx.base import SphinxBase, Group
from pyiron_atomistics.atomistics.job.interactive import (
    GenericInteractive,
    GenericInteractiveOutput,
)
from pyiron_atomistics.vasp.potential import VaspPotentialSetter

BOHR_TO_ANGSTROM = (
    scipy.constants.physical_constants["Bohr radius"][0] / scipy.constants.angstrom
)
HARTREE_TO_EV = scipy.constants.physical_constants["Hartree energy in eV"][0]
HARTREE_OVER_BOHR_TO_EV_OVER_ANGSTROM = HARTREE_TO_EV / BOHR_TO_ANGSTROM

__author__ = "Osamu Waseda, Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class SphinxInteractive(SphinxBase, GenericInteractive):
    def __init__(self, project, job_name):
        super(SphinxInteractive, self).__init__(project, job_name)
        self.output = SphinxOutput(job=self)
        self._interactive_write_input_files = True
        self._interactive_library_read = None
        self._interactive_fetch_completed = True
        self.interactive_flush_frequency = 1

    @property
    def structure(self):
        return GenericInteractive.structure.fget(self)

    @structure.setter
    def structure(self, structure):
        GenericInteractive.structure.fset(self, structure)
        if structure is not None:
            self._potential = VaspPotentialSetter(
                element_lst=structure.get_species_symbols().tolist()
            )

    def _get_structure(self, frame=-1, wrap_atoms=True):
        return GenericInteractive._get_structure(
            self, frame=frame, wrap_atoms=wrap_atoms
        )

    def interactive_energy_tot_getter(self):
        return self.interactive_energy_pot_getter()

    def interactive_energy_pot_getter(self):
        self._interactive_pipe_write("get energy")
        return float(self._interactive_library_read.readline()) * HARTREE_TO_EV

    def interactive_forces_getter(self):
        self._interactive_pipe_write("get forces")
        ff = []
        for _ in range(len(self.structure)):
            line = self._interactive_library_read.readline().split()
            ff.append(
                [
                    float(line[i]) * HARTREE_OVER_BOHR_TO_EV_OVER_ANGSTROM
                    for i in range(3)
                ]
            )
        ff = np.array(ff)[self.id_spx_to_pyi]
        return ff

    def interactive_cells_getter(self):
        self._interactive_pipe_write("get cell")
        cc = []
        for _ in range(3):
            line = self._interactive_library_read.readline().split()
            cc.append([float(line[i]) * BOHR_TO_ANGSTROM for i in range(3)])
        return np.array(cc)

    def interactive_positions_getter(self):
        self._interactive_pipe_write("get structure")
        xx = []
        for _ in range(len(self.structure)):
            line = self._interactive_library_read.readline().split()
            xx.append([float(line[i]) * BOHR_TO_ANGSTROM for i in range(3)])
        xx = np.array(xx)[self.id_spx_to_pyi]
        return xx

    def interactive_positions_setter(self, positions):
        self._interactive_pipe_write("set structure")
        positions = positions[self.id_pyi_to_spx]
        positions = np.reshape(positions, 3 * len(self.structure)) / BOHR_TO_ANGSTROM
        self._interactive_pipe_write(positions.tolist())

    def interactive_spins_getter(self):
        self._logger.debug("get spins - start ...")
        self._interactive_pipe_write("get atomspin")
        mm = []
        for _ in range(len(self.structure)):
            line = self._interactive_library_read.readline().split()
            mm.append(float(line[0]))
        mm = np.array(mm)[self.id_spx_to_pyi]
        # self.interactive_cache['atom_spins'].append(mm)
        self._logger.debug("get spins - done.")
        return mm

    def interactive_spin_constraints_setter(self, spins):
        if self._generic_input["fix_spin_constraint"]:
            self._logger.debug("set spin constraints - start ...")
            self._spin_constraint_enabled = True
            self._interactive_pipe_write("set spinconstraint")
            spins = np.array(spins)[self.id_pyi_to_spx]
            self._spin_constraints = np.array(spins)
            self._interactive_pipe_write(spins.tolist())
            # self.interactive_cache['atom_spin_constraints'].append(spins)
            self._logger.debug("set spin constraints - done.")
        else:
            warnings.warn("Spin constraint not set -> set fix_spin_constraint = True")

    def interactive_spin_constraints_getter(self):
        return self._spin_constraints
        # return self.interactive_cache['atom_spin_constraints'][-1]

    def interactive_magnetic_forces_getter(self):
        if self._generic_input["fix_spin_constraint"]:
            self._interactive_pipe_write("get nu")
            nn = []
            for _ in range(len(self.structure)):
                line = self._interactive_library_read.readline().split()
                nn.append(HARTREE_TO_EV * float(line[0]))
            nn = np.array(nn)[self.id_spx_to_pyi]
            return nn
        else:
            return None

    def interactive_initialize_interface(self):
        self.server.threads = self.input["THREADS"]
        if self.executable.executable_path == "":
            self.status.aborted = True
            raise ValueError("No executable set!")
        if self.server.cores == 1 or not self.executable.mpi:
            out = subprocess.Popen(
                str(self.executable),
                cwd=self.project_hdf5.working_directory,
                shell=True,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
        else:
            out = subprocess.Popen(
                [
                    self.executable.executable_path,
                    str(self.server.cores),
                    str(self.server.threads),
                ],
                cwd=self.project_hdf5.working_directory,
                shell=False,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
        while not self._interactive_pipes_initialized:
            time.sleep(1)
        self._logger.debug("open interactive interface!")
        self._interactive_library = open(
            os.path.join(self.working_directory, "sxctrl"), "w"
        )
        self._interactive_library_read = open(
            os.path.join(self.working_directory, "sxres"), "r"
        )
        self._logger.debug("interactive interface is opened!")
        if (
            not self.structure.has("initial_magmoms")
            and "atom_spins" in self.interactive_cache.keys()
        ):
            del self.interactive_cache["atom_spins"]
        if self._generic_input["fix_spin_constraint"]:
            self.interactive_spin_constraints_setter(
                self._structure_current.get_initial_magnetic_moments()
            )
        else:
            if "magnetic_forces" in self.interactive_cache.keys():
                del self.interactive_cache["magnetic_forces"]
            if "atom_spin_constraints" in self.interactive_cache.keys():
                del self.interactive_cache["atom_spin_constraints"]
        if len(self.restart_file_list) > 0:
            self._logger.debug("restarting; interactive run - start ...")
            self._interactive_pipe_write("run restart")
            self.interactive_fetch()

    def _output_interactive_to_generic(self):
        with self.project_hdf5.open("output") as h5:
            if "interactive" in h5.list_groups():
                for key in ["positions", "cells", "indices", "cells", "forces"]:
                    self._output_parser.generic[key] = h5["interactive/" + key]
                if "atom_spin_constraints" in h5["interactive"].list_nodes():
                    self._output_parser.generic.dft.atom_spin_constraints = h5[
                        "interactive/atom_spin_constraints"
                    ]
                self._output_parser.generic.to_hdf(hdf=self.project_hdf5)

    def collect_output(self, force_update=False, compress_files=True):
        super(SphinxInteractive, self).collect_output(
            force_update=force_update, compress_files=compress_files
        )
        self._output_interactive_to_generic()

    def interactive_close(self):
        if self.interactive_is_activated():
            self._interactive_pipe_write("end")
            self._interactive_library.close()
            self._interactive_library_read.close()
            self.status.collect = True
            if self["energy.dat"] is not None:
                self.run()
            self._output_interactive_to_generic()
            super(SphinxInteractive, self).interactive_close()

    def calc_minimize(
        self,
        electronic_steps=None,
        ionic_steps=None,
        max_iter=None,
        pressure=None,
        algorithm=None,
        retain_charge_density=False,
        retain_electrostatic_potential=False,
        ionic_energy_tolerance=None,
        ionic_force_tolerance=None,
        ionic_energy=None,
        ionic_forces=None,
        volume_only=False,
    ):
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            raise NotImplementedError(
                "calc_minimize() is not implemented for the interactive mode use calc_static()!"
            )
        else:
            super(SphinxInteractive, self).calc_minimize(
                electronic_steps=electronic_steps,
                ionic_steps=ionic_steps,
                max_iter=max_iter,
                pressure=pressure,
                algorithm=algorithm,
                retain_charge_density=retain_charge_density,
                retain_electrostatic_potential=retain_electrostatic_potential,
                ionic_energy_tolerance=ionic_energy_tolerance,
                ionic_force_tolerance=ionic_force_tolerance,
                volume_only=volume_only,
            )

    def run_if_interactive(self):
        super(SphinxInteractive, self).run_if_interactive()
        self._logger.debug("interactive run - start ...")
        self._interactive_pipe_write("run electronicminimization")
        self.interactive_fetch()

    def run_if_interactive_non_modal(self):
        if not self._interactive_fetch_completed:
            print("Warning: interactive_fetch being effectuated")
            self.interactive_fetch()
        super(SphinxInteractive, self).run_if_interactive()
        self._logger.debug("interactive run - start ...")
        self._interactive_pipe_write("run electronicminimization")
        self._interactive_fetch_completed = False

    def interactive_fetch(self):
        if (
            self._interactive_fetch_completed
            and self.server.run_mode.interactive_non_modal
        ):
            print("First run and then fetch")
        else:
            self.interactive_collect()
            self._logger.debug("interactive run - done")

    @property
    def _interactive_pipes_initialized(self):
        return os.path.exists(
            os.path.join(self.working_directory, "sxctrl")
        ) and os.path.exists(os.path.join(self.working_directory, "sxres"))

    def _interactive_pipe_write(self, line):
        if isinstance(line, str) or isinstance(line, int) or isinstance(line, float):
            self._interactive_library.write(str(line) + "\n")
            self._interactive_library.flush()
        elif isinstance(line, list):
            for subline in line:
                self._interactive_pipe_write(subline)
        else:
            raise TypeError("only lists, strings and numbers are supported!")

    def _interactive_pipe_read(self):
        return self._interactive_library_read.readline()

    def calc_static(
        self,
        electronic_steps=100,
        blockSize=8,
        dSpinMoment=1e-8,
        algorithm=None,
        retain_charge_density=False,
        retain_electrostatic_potential=False,
    ):
        """
        Function to setup the hamiltonian to perform static SCF DFT runs

        Args:
            retain_electrostatic_potential:
            retain_charge_density:
            algorithm:
            electronic_steps (int): maximum number of electronic steps, which can be used to achieve convergence
        """
        super(SphinxInteractive, self).calc_static(
            electronic_steps=electronic_steps,
            algorithm=algorithm,
            retain_charge_density=retain_charge_density,
            retain_electrostatic_potential=retain_electrostatic_potential,
        )

    def load_main_group(self):
        main_group = Group()
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            commands = Group(
                [
                    {
                        "id": '"restart"',
                        "scfDiag": self.get_scf_group(
                            maxSteps=10, keepRhoFixed=True, dEnergy=1.0e-4
                        ),
                    },
                    {
                        "id": '"electronicminimization"',
                        "scfDiag": self.get_scf_group(),
                    },
                ]
            )
            self.input.sphinx.main.extControl = Group()
            self.input.sphinx.main.extControl.set_group("bornOppenheimer")
            self.input.sphinx.main.extControl.bornOppenheimer = commands
        else:
            super(SphinxInteractive, self).load_main_group()


class SphinxOutput(GenericInteractiveOutput):
    def __init__(self, job):
        super(SphinxOutput, self).__init__(job)

    def check_band_occupancy(self, plot=True):
        """
        Check whether there are still empty bands available.

        args:
            plot (bool): plots occupancy of the last step

        returns:
            True if there are still empty bands
        """
        import matplotlib.pylab as plt

        elec_dict = self._job["output/generic/dft"]["n_valence"]
        if elec_dict is None:
            raise AssertionError("Number of electrons not parsed")
        n_elec = np.sum(
            [elec_dict[k] for k in self._job.structure.get_chemical_symbols()]
        )
        n_elec = int(np.ceil(n_elec / 2))
        bands = self._job["output/generic/dft/bands_occ"][-1]
        bands = bands.reshape(-1, bands.shape[-1])
        max_occ = np.sum(~np.isclose(bands, 0), axis=-1).max()
        n_bands = bands.shape[-1]
        if plot:
            xticks = np.arange(1, n_bands + 1)
            plt.xlabel("Electron number")
            plt.ylabel("Occupancy")
            if n_bands < 20:
                plt.xticks(xticks)
            plt.axvline(n_elec, label="#electrons: {}".format(n_elec))
            plt.axvline(max_occ, color="red", label="Max occupancy: {}".format(max_occ))
            plt.axvline(
                n_bands, color="green", label="Number of bands: {}".format(n_bands)
            )
            plt.plot(xticks, bands.T, "x", color="black")
            plt.legend()
        if max_occ < n_bands:
            return True
        else:
            return False


class SphinxInt2(SphinxInteractive):
    def __init__(self, project, job_name):
        warnings.warn("Please use SphinxInt instead of SphinxInt2")
        super(SphinxInt2, self).__init__(project=project, job_name=job_name)



__author__ = "Christoph Freysoldt"
__copyright__ = (
    "Copyright 2022, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Christoph Freysoldt"
__email__ = "freysoldt@mpie.de"
__status__ = "development"
__date__ = "Oct 15, 2022"

from threading import Lock
import json
import warnings
import os.path
from sys import executable as python_interpreter
import subprocess

from pyiron_base import Settings

_sxversion_lock = Lock()
_sxversions = None


def sxversions(refresh=False):
    """Get available SPHInX versions
    returns: a dict
    """
    global _sxversions
    if not refresh and _sxversions is dict:
        return _sxversions

    def do_update(warn_what, newfile):
        if isinstance(sxv, dict):
            for v in sxv.keys():
                if v in version_origin.keys() and _sxversions[v] != sxv[v]:
                    warnings.warn(
                        "Overriding sxversion '{}' from {} with the one from {}".format(
                            v, version_origin[v], newfile
                        )
                    )
                version_origin[v] = newfile
            _sxversions.update(sxv)
        else:
            warnings.warn("Failed to parse " + warn_what + newfile)

    _sxversion_lock.acquire()
    try:
        if refresh:
            _sxversions = None
        if _sxversions is None:
            _sxversions = dict()
            version_origin = dict()
            for p in Settings().resource_paths:
                jsonfile = os.path.join(p, "sphinx", "sxversions.json")
                if os.path.exists(jsonfile):
                    with open(jsonfile) as f:
                        sxv = json.load(f)
                    do_update("", jsonfile)

                jsonscript = os.path.join(p, "sphinx", "sxversions.py")
                if os.path.exists(jsonscript):
                    proc = subprocess.run(
                        [python_interpreter, jsonscript],
                        text=True,
                        stdout=subprocess.PIPE,
                        cwd=os.path.join(p, "sphinx"),
                    )
                    if proc.returncode != 0:
                        warnings.warn(
                            jsonscript
                            + " failed with exitcode "
                            + str(proc.returncode)
                            + ": \n"
                            + proc.stdout
                            + proc.stderr
                        )
                    else:
                        try:
                            sxv = json.loads(proc.stdout)
                            do_update("output from ", jsonscript)
                        except json.decoder.JSONDecodeError as ex:
                            raise RuntimeError(
                                "json decoder error from "
                                + jsonscript
                                + " output:\n"
                                + proc.stdout
                            ) from ex
    finally:
        _sxversion_lock.release()
    return _sxversions



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from abc import ABC, abstractmethod, abstractproperty
from typing import Callable
import numbers
from pyiron_base import deprecate, ImportAlarm
from pyiron_atomistics.atomistics.structure.atoms import Atoms

with ImportAlarm("Animation of atomic structures requires nglview") as nglview_alarm:
    import nglview

"""
Mixin for classes that have one or more structures attached to them.
"""

__author__ = "Marvin Poul"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.2"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "production"
__date__ = "Apr 22, 2021"


class HasStructure(ABC):
    """
    Mixin for classes that have one or more structures attached to them.

    Necessary overrides are :meth:`._get_structure()` and
    :meth:`._number_of_structures()`.

    :meth:`.get_structure()` checks that iteration_step is valid; implementations of
    :meth:`._get_structure()` therefore don't have to check it.

    :attr:`.number_of_structures` may be zero, e.g. if there's no structure stored in the object yet or a
    job will compute this structure, but hasn't been run yet.

    Sub classes that wish to document special behavior of their implementation of :meth:`.get_structure` may do so by
    adding documention to it in the "Methods:" sub section of their class docstring.

    Sub classes may support custom data types as indices for `frame` in :meth:`.get_structure()` by overriding
    :meth:`._translate_frame()`.

    The example below shows how to implement this mixin and how to check whether an object derives from it

    >>> from pyiron_atomistics.atomistics.structure.atoms import Atoms
    >>> class Foo(HasStructure):
    ...     '''
    ...     Methods:
    ...         .. method:: get_structure
    ...             returns structure with single Fe atom at (0, 0, 0)
    ...     '''
    ...     def _get_structure(self, frame=-1, wrap_atoms=True):
    ...         return Atoms(symbols=['Fe'], positions=[[0,0,0]])
    ...     def _number_of_structures(self):
    ...         return 1

    >>> f = Foo()
    >>> for s in f.iter_structures():
    ...     print(s)
    Fe: [0. 0. 0.]
    pbc: [False False False]
    cell:
    Cell([0.0, 0.0, 0.0])
    <BLANKLINE>

    >>> isinstance(f, HasStructure)
    True


    .. document private functions
    .. automethod:: _get_structure
    .. automethod:: _number_of_structures
    .. automethod:: _translate_frame
    """

    @deprecate(iteration_step="use frame instead")
    def get_structure(self, frame=-1, wrap_atoms=True, iteration_step=None):
        """
        Retrieve structure from object.  The number of available structures depends on the job and what kind of
        calculation has been run on it, see :attr:`.number_of_structures`.

        Args:
            frame (int, object): index of the structure requested, if negative count from the back; if
            :meth:`_translate_frame()` is overridden, frame will pass through it
            iteration_step (int): deprecated alias for frame
            wrap_atoms (bool): True if the atoms are to be wrapped back into the unit cell

        Returns:
            :class:`pyiron_atomistics.atomistics.structure.atoms.Atoms`: the requested structure

        Raises:
            IndexError: if not -:attr:`.number_of_structures` <= iteration_step < :attr:`.number_of_structures`
        """
        if iteration_step is not None:
            frame = iteration_step
        if not isinstance(frame, numbers.Integral):
            try:
                frame = self._translate_frame(frame)
            except NotImplementedError:
                raise KeyError(
                    f"argument frame {frame} is not an integer and _translate_frame() not implemented!"
                ) from None
        num_structures = self.number_of_structures
        if frame < 0:
            frame += num_structures
        if not (0 <= frame < num_structures):
            raise IndexError(
                f"argument frame {frame} out of range [-{num_structures}, {num_structures})."
            )

        return self._get_structure(frame=frame, wrap_atoms=wrap_atoms)

    def _translate_frame(self, frame):
        """
        Translate frame to an integer for :meth:`_get_structure()`.

        Args:
            frame (object): any object to translate into an integer id

        Returns:
            int: valid integer to be passed to :meth:`._get_structure()`

        Raises:
            KeyError: if given frame does not exist in this object
        """
        raise NotImplementedError("No frame translation implemented!")

    @abstractmethod
    def _get_structure(self, frame=-1, wrap_atoms=True):
        pass

    @property
    def number_of_structures(self):
        """
        `int`: maximum `iteration_step` + 1 that can be passed to :meth:`.get_structure()`.
        """
        return self._number_of_structures()

    @abstractmethod
    def _number_of_structures(self):
        pass

    def iter_structures(self, wrap_atoms=True):
        """
        Iterate over all structures in this object.

        Args:
            wrap_atoms (bool): True if the atoms are to be wrapped back into the unit cell; passed to
                               :meth:`.get_structure()`

        Yields:
            :class:`pyiron_atomistics.atomistitcs.structure.atoms.Atoms`: every structure attached to the object
        """
        for i in range(self.number_of_structures):
            yield self._get_structure(frame=i, wrap_atoms=wrap_atoms)

    def transform_structures(self, modify) -> "TransformStructure":
        """
        Return a modified object by applying a function to each object lazily.

        Args:
            modify (function): applied to each structure, has to return the modified structure

        Returns:
            :class:`.TransformStructure`: a container with the modified structures
        """
        return TransformStructure(self, modify)

    def collect_structures(self, filter_function=None) -> "StructureStorage":
        """
        Collects a copy of all structures in a compact :class:`.StructureStorage`.

        This can be used to force lazily applied modifications with :meth:`.transform_structures` or simply to obtain a
        known object type from a generic :class:`.HasStructure` object.

        Args:
            filter_function (function): include structure only if this function returns True for it

        Returns:
            :class:`.StructureStorage`: a copy of all (filtered) structures
        """
        # breaks cyclical import
        # this is a bit annoying, but I want to give users an entry point to using StructureStorage without having to
        # import it
        from pyiron_atomistics.atomistics.structure.structurestorage import (
            StructureStorage,
        )

        store = StructureStorage()
        for structure in self.iter_structures():
            if filter_function is None or filter_function(structure):
                store.add_structure(structure)
        return store

    @nglview_alarm
    def animate_structures(
        self,
        spacefill: bool = True,
        show_cell: bool = True,
        center_of_mass: bool = False,
        particle_size: float = 0.5,
        camera: str = "orthographic",
    ):
        """
        Animate a series of atomic structures.

        Args:
            spacefill (bool): If True, then atoms are visualized in spacefill stype
            show_cell (bool): True if the cell boundaries of the structure is to be shown
            particle_size (float): Scaling factor for the spheres representing the atoms.
                                    (The radius is determined by the atomic number)
            center_of_mass (bool): False (default) if the specified positions are w.r.t. the origin
            camera (str): camera perspective, choose from "orthographic" or "perspective"

        Returns:
            animation: nglview IPython widget
        """

        if self._number_of_structures() <= 1:
            raise ValueError("job must have more than one structure to animate!")

        animation = nglview.show_asetraj(_TrajectoryAdapter(self))
        if spacefill:
            animation.add_spacefill(radius_type="vdw", scale=0.5, radius=particle_size)
            animation.remove_ball_and_stick()
        else:
            animation.add_ball_and_stick()
        if show_cell:
            animation.add_unitcell()
        animation.camera = camera
        return animation


class TransformStructure(HasStructure):
    """
    Modifies any HasStructure by applying a function to each structure lazily.
    """

    __slots__ = ("_source", "_modify")

    def __init__(self, source: HasStructure, modify: Callable[[Atoms], Atoms]):
        self._source = source
        self._modify = modify

    def _number_of_structures(self):
        return self._source._number_of_structures()

    def _translate_frame(self, frame):
        return self._source._translate_frame(frame)

    def _get_structure(self, frame=-1, wrap_atoms=True):
        return self._modify(self._source._get_structure(frame, wrap_atoms=wrap_atoms))


class _TrajectoryAdapter:
    """
    Class that translates between HasStructure and the ASE Trajectory interface.

    The ASE interface is needed e.g. by nglview to animate a series of structures, but the Trajectory interface uses
    methods that not all can implement (e.g. AtomisticGenericJob already overloads __getitem__).
    """

    __slots__ = "_underlying"

    def __init__(self, underlying: HasStructure):
        self._underlying = underlying

    def __getitem__(self, item):
        return self._underlying.get_structure(item)

    def __len__(self):
        return self._underlying.number_of_structures

    def __iter__(self):
        yield from self._underlying.iter_structures()
